/*
 * Turbo Programmer Utilities, turbo-prog-utils, www.bladox.com 
 *
 * Copyright (C) 2004 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <tfslib.h>
#include <tproglib.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "../module-src/df.h"

void help ()
{
  printf ("Remove file or directory\n");
  printf ("Usage: [args] inode_in_hex\n");
  printf ("-h\thelp\n");
  printf ("-d device\tuse USB device\n");
  printf ("-r\trecursive\n");
  printf ("-s\tfiles in dir\n");
  printf ("-v\tverbose\n");
  exit (-1);
}

int main (int argc, char **argv)
{
  u32 i;
  u16 c;
  u8 nb;
  u16 pos;
  u8 buf[1000];
  u8 len;
  int result;
  u8 sort;
  u16 j;
  u8 *d_name;
  u8 path[1000] = { '/', '\0' };
  u8 name[1000];
  u16 name_end;
  u16 path_end;
  b32 f;
  u8 rec = 0;
  Inode ino;
  u8 o_s = 0;
  u8 *dev = USBDEVICE;

  do
  {
    result = getopt (argc, argv, "rhsd:v");
    switch (result)
    {
      case 'd':
	dev = optarg;
	break;
      case 's':
	o_s = 1;
	break;
      case 'r':
	rec = 1;
	break;
      case 'h':
	help ();
	break;
      case 'v':
	verbose_inc ();
	break;
    }
  }
  while (result > 0);

  if (argc - optind != 1)
    help ();
  sscanf (argv[optind], "%x", &i);
  f = fs_ino_p (i >> 8, i & 0xFF);
//printf("DST %x %04X %02X\n", f, fs_ino_block(f), fs_ino_index(f));

  open_device (dev);
  if (select_app (APP_DF) != NO_ERROR)
  {
    close_device ();
    fprintf (stderr, "Cannot open application %s\n", APP_DF);
    exit (-1);
  }

  usb_send (ACTION_DF_INIT, 0, NULL, NULL, NULL);

  if (mc_detect () == MC_OK)
  {
    u8 name[100];

    printf ("Memory card detected\n");
    mc_get_label (name);
    printf ("Card label: %s\n\n", name);
  }
  else if (mc_detect () == MC_NOT_PRESENT)
  {
    printf ("Memory card NOT PRESENT\n");
    close_device ();
    exit (1);
  }
  else if (mc_detect () == MC_NOT_READY)
  {
    printf ("Memory card NOT FORMATED\n\n");
    exit (1);
  }

  get_inode (&ino, f);
  printf ("Inode: %04X%02X\n", fs_ino_block (f), fs_ino_index (f));

  if (ino.type == INODE_TYPE_DIR && rec == 0 && o_s == 0)
  {
    fprintf (stderr, "%04X%02X is a directory, use -r\n", fs_ino_block (f),
	     fs_ino_index (f));
    goto X1;
  }

  if (o_s && ino.type != INODE_TYPE_DIR)
  {
    fprintf (stderr, "%04X%02X not a directory\n", fs_ino_block (f),
	     fs_ino_index (f));
    goto X1;
  }

  if (o_s)
    delete_files (f);
  else
    delete_inode (f);

X1:
  close_device ();
}
