/*
 * Turbo Programmer Utilities, turbo-prog-utils, www.bladox.com 
 *
 * Copyright (C) 2004 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <tprog/tproglib.h>
#include "../module-src/turboadapter.h"

#ifdef WIN32_MINGW
#include <windows.h>
#endif

void help ()
{
  printf ("Print Turbo Adapter debug\n");
  printf ("Usage: btldr [args] file\n");
  printf ("-a application\tuse application, default is %s\n",
	  APP_TURBO_ADAPTER);
  printf ("-d device\tuse USB device, default is %s\n", USBDEVICE);
  printf ("-e file\t\tupload EEPROM image too\n");
  printf ("-s\t\tuse SIM connector\n");
  printf ("-h\t\thelp\n");
  printf ("-v\t\tverbose\n");
  printf ("-t\t\ttest, verify content of memory\n");
  exit (-1);
}

int main (int argc, char **argv)
{
  int result;
  u8 *dev = USBDEVICE;
  u8 *app_name = APP_TURBO_ADAPTER;
  u8 answer[2048];
  u16 answerLen;
  u32 i, i2;
  u16 l;
  u16 flashPageNumber;
  int fdstdin;
  int flags;
  int bytes_read;
  u8 buf[1024];
  u8 verbuf[1024];
  int len;
  FILE *fe;
  u8 *filename = NULL;
  u8 *efile = NULL;
  u8 c;
  u8 buf2[0x20000];
  u8 buf3[0x2000];
  u8 use_sim = 0;
  u8 verify = 0;
  u8 was_error = 0;

  do
  {
    result = getopt (argc, argv, "hd:a:vse:t");
    switch (result)
    {
      case 'a':
	app_name = optarg;
	break;
      case 'd':
	dev = optarg;
	break;
      case 'e':
	efile = optarg;
	break;
      case 's':
	use_sim = 1;
	break;
      case 't':
	verify = 1;
	break;
      case 'v':
	verbose_inc ();
	break;
      case 'h':
	help ();
	break;
    }
  }
  while (result > 0);

  if (argc - optind != 1)
    help ();

  filename = argv[optind];

#ifndef WIN32_MINGW
  fe = fopen (filename, "rb");
  if (fe == NULL)
  {
    printf ("Cannot open %s for reading\n", filename);
    exit (-1);
  }

  for (i2 = 0; i2 < 0x1E000; i2++)
    buf2[i2] = 0xFF;

  i2 = 0;
  while (!feof (fe))
  {
    buf2[i2] = fgetc (fe);
    i2++;
  }
  fclose (fe);

  if (efile)
  {
    fe = fopen (efile, "rb");
    if (fe == NULL)
    {
      printf ("Cannot open %s for reading\n", efile);
      exit (-1);
    }

    for (i2 = 0; i2 < 0x1000; i2++)
      buf3[i2] = 0xFF;

    i2 = 0;
    while (!feof (fe))
    {
      buf3[i2] = fgetc (fe);
      i2++;
    }
    fclose (fe);
  }
#else
  HANDLE *f1;
  int n_in;

  f1 =
    CreateFile (argv[optind], GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);

  if (f1 == INVALID_HANDLE_VALUE)
  {
    printf ("Cannot open file %s (error %d)\n", argv[optind],
	    GetLastError ());
    exit (-1);
  }

  for (i2 = 0; i2 < 0x1E000; i2++)
    buf2[i2] = 0xFF;

  ReadFile (f1, buf2, 0x1E000, (DWORD *) (&n_in), NULL);

  CloseHandle (f1);

  if (efile)
  {
    f1 = CreateFile (efile, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);

    if (f1 == INVALID_HANDLE_VALUE)
    {
      printf ("Cannot open file %s (error %d)\n", efile, GetLastError ());
      exit (-1);
    }

    for (i2 = 0; i2 < 0x1000; i2++)
      buf3[i2] = 0xFF;

    ReadFile (f1, buf3, 0x1000, (DWORD *) (&n_in), NULL);

    CloseHandle (f1);
  }
#endif

  open_device (dev);
  if (select_app (app_name) != NO_ERROR)
  {
    close_device ();
    fprintf (stderr, "Cannot open application %s\n", app_name);
    exit (-1);
  }

  printf ("Turbo Adapter bootloader mode initialized, %s\n", app_name);
  printf ("BOOTLOADER is erasing FLASH and EEPROM, wait for approx. 50s\n");
  if (use_sim)
    usb_send (ACTION_TURBO_ADAPTER_BTLDR_SIM_INIT, 0, NULL, &answerLen,
	      answer);
  else
    usb_send (ACTION_TURBO_ADAPTER_BTLDR_INIT, 0, NULL, &answerLen, answer);

  printf ("BOOTLOADER is updating FLASH\n");
  for (i2 = 0; i2 < 0x1E000; i2 += 256)
  {
    printf ("Page %4d/(0-%4d)\r", i2 / 256, 0x1E0 - 1);
    fflush (stdout);
    flashPageNumber = i2 / 256;	//address(last page before 4kB bootloader)= 0x1DF00 bytes(0xEF00 words) (page=256B)
    answer[0] = (flashPageNumber & 0xFF);	//low byte
    answer[1] = (flashPageNumber >> 8);	//high byte
    for (i = 0; i < 256; i++)
    {
      answer[2 + i] = buf2[i2 + i];
    }
    usb_send (ACTION_TURBO_ADAPTER_BTLDR_WRITE_PAGE, 2 + 256, answer,
	      &answerLen, answer);

    if (verify)
    {
      answer[0] = (flashPageNumber & 0xFF);	//low byte
      answer[1] = (flashPageNumber >> 8);	//high byte
      usb_send (ACTION_TURBO_ADAPTER_BTLDR_READ_PAGE, 2, answer,
		&answerLen, verbuf);
      if (answerLen != 256)
      {
	printf ("ERROR: verify wrong answer len %d\n", answerLen);
	was_error = 1;
      }

      for (i = 0; i < answerLen; i++)
      {
	if (verbuf[i] != buf2[i2 + i])
	{
	  printf ("ERROR: verify FLASH at pos %06X: %02X, should be %02X\n",
		  i2 + i, verbuf[i], buf2[i2 + i]);
	  was_error = 1;
	}
      }
    }
  }
  printf ("\n");

  if (efile)
  {
    printf ("BOOTLOADER is updating EEPROM\n");
    for (i2 = 0; i2 < 0x1000; i2 += 8)
    {
      printf ("Page %4d/(0-%4d)\r", i2 / 8, 511);
      fflush (stdout);
      flashPageNumber = i2 / 8;	//address page len 8b
      answer[0] = (flashPageNumber & 0xFF);	//low byte
      answer[1] = (flashPageNumber >> 8);	//high byte
      for (i = 0; i < 8; i++)
	answer[2 + i] = buf3[i2 + i];
      usb_send (ACTION_TURBO_ADAPTER_BTLDR_WRITE_PAGE_EEPROM, 2 + 8, answer,
		&answerLen, answer);

      if (verify)
      {
	answer[0] = (flashPageNumber & 0xFF);	//low byte
	answer[1] = (flashPageNumber >> 8);	//high byte
	usb_send (ACTION_TURBO_ADAPTER_BTLDR_READ_PAGE_EEPROM, 2, answer,
		  &answerLen, verbuf);
	if (answerLen != 8)
	{
	  printf ("ERROR: verify wrong answer len %d\n", answerLen);
	  was_error = 1;
	}

	for (i = 0; i < answerLen; i++)
	{
	  if (verbuf[i] != buf3[i2 + i])
	  {
	    printf
	      ("ERROR: verify EEPROM at pos %04X: %02X, should be %02X\n",
	       i2 + i, verbuf[i], buf3[i2 + i]);
	    was_error = 1;
	  }
	}
      }
    }
    printf ("\n");
  }

  close_device ();
  if (was_error)
  {
    printf ("ERROR!\n");
    exit (-1);
  }
}
